@@ -32,7 +32,8 @@ class LocalMetadata():
             if "dir" in func: continue
             if "open" in func or "close" in func or "creat" in func \
                 or "seek" in func or "sync" in func:
-                self.filemap.add(r.args[0] if type(r.args[0])==str else r.args[i].decode('utf-8'))
+                fstr = r.args[0]
+                self.filemap.add(fstr if type(fstr)==str else fstr.decode('utf-8'))
 
         self.num_files = len(self.filemap)
 
