@@ -5,26 +5,36 @@
 #include <assert.h>
 #include "./reader.h"
 
-static int mpi_start_idx = -1;
-static int hdf5_start_idx = -1;
+void check_version(RecorderReader* reader) {
+    char version_file[1024];
+    snprintf(version_file, sizeof(version_file), "%s/VERSION", reader->logs_dir);
 
-static double prev_tstart = 0;
-
-void read_metadata(char* path, RecorderMetadata *metadata) {
-    FILE* fp = fopen(path, "rb");
+    FILE* fp = fopen(version_file, "r");
     assert(fp != NULL);
-    fread(metadata, sizeof(RecorderMetadata), 1, fp);
+    int major, minor, patch;
+    fscanf(fp, "%d.%d.%d", &major, &minor, &patch);
+    if(major != VERSION_MAJOR || minor != VERSION_MINOR) {
+        fprintf(stderr, "incompatible version: file=%d.%d.%d != reader=%d.%d.%d\n",
+                major, minor, patch, VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH);
+        exit(1);
+    }
     fclose(fp);
 }
 
-void read_func_list(char* path, RecorderReader *reader) {
-    FILE* fp = fopen(path, "rb");
+void read_metadata(RecorderReader* reader) {
+    char metadata_file[1024];
+    snprintf(metadata_file, sizeof(metadata_file), "%s/recorder.mt", reader->logs_dir);
 
+    FILE* fp = fopen(metadata_file, "rb");
+    assert(fp != NULL);
+    fread(&reader->metadata, sizeof(reader->metadata), 1, fp);
+
+    long pos = ftell(fp);
     fseek(fp, 0, SEEK_END);
-    long fsize = ftell(fp) - sizeof(RecorderMetadata);
+    long fsize = ftell(fp) - pos;
     char buf[fsize];
 
-    fseek(fp, sizeof(RecorderMetadata), SEEK_SET); // skip RecorderMetadata object
+    fseek(fp, pos, SEEK_SET);
     fread(buf, 1, fsize, fp);
 
     int start_pos = 0, end_pos = 0;
@@ -35,14 +45,14 @@ void read_func_list(char* path, RecorderReader *reader) {
             memset(reader->func_list[func_id], 0, sizeof(reader->func_list[func_id]));
             memcpy(reader->func_list[func_id], buf+start_pos, end_pos-start_pos);
             start_pos = end_pos+1;
-            if((mpi_start_idx==-1) &&
+            if((reader->mpi_start_idx==-1) &&
                 (NULL!=strstr(reader->func_list[func_id], "MPI")))
-                mpi_start_idx = func_id;
+                reader->mpi_start_idx = func_id;
 
-            if((hdf5_start_idx==-1) &&
+            if((reader->hdf5_start_idx==-1) &&
                 (NULL!=strstr(reader->func_list[func_id], "H5")))
 
-                hdf5_start_idx = func_id;
+                reader->hdf5_start_idx = func_id;
 
             func_id++;
         }
@@ -52,15 +62,66 @@ void read_func_list(char* path, RecorderReader *reader) {
 }
 
 void recorder_init_reader(const char* logs_dir, RecorderReader *reader) {
-    char metadata_file[1024];
-    strcpy(reader->logs_dir, logs_dir);
+    assert(logs_dir);
+    assert(reader);
 
-    sprintf(metadata_file, "%s/recorder.mt", logs_dir);
-    read_metadata(metadata_file, &reader->metadata);
-    read_func_list(metadata_file, reader);
+    memset(reader, 0, sizeof(*reader));
+    strcpy(reader->logs_dir, logs_dir);
+    reader->mpi_start_idx = -1;
+    reader->hdf5_start_idx = -1;
+    reader->prev_tstart = 0.0;
+
+    check_version(reader);
+
+    read_metadata(reader);
+
+	int nprocs= reader->metadata.total_ranks;
+
+	reader->ug_ids = malloc(sizeof(int) * nprocs);
+	reader->csts   = malloc(sizeof(CST*) * nprocs);
+	reader->cfgs   = malloc(sizeof(CFG*) * nprocs);
+
+	for(int i = 0; i < nprocs; i++) {
+		reader->ug_ids[i] = i;
+		reader->csts[i] = NULL;
+		reader->cfgs[i] = NULL;
+	}
+
+	if(reader->metadata.interprocess_compression) {
+		recorder_read_cst(reader, 0);
+		for(int i = 1; i < nprocs; i++)
+			reader->csts[i] = reader->csts[0];
+
+		char ug_metadata_fname[1024] = {0};
+		sprintf(ug_metadata_fname, "%s/ug.mt", reader->logs_dir);
+		FILE* f = fopen(ug_metadata_fname, "rb");
+		fread(reader->ug_ids, sizeof(int), nprocs, f);
+		fread(&reader->num_ugs, sizeof(int), 1, f);
+		fclose(f);
+
+		for(int i = 0; i < reader->num_ugs; i++) {
+			recorder_read_cfg(reader, i);
+		}
+	}
 }
 
 void recorder_free_reader(RecorderReader *reader) {
+    assert(reader);
+	free(reader->ug_ids);
+
+	if(reader->metadata.interprocess_compression) {
+		recorder_free_cst(reader->csts[0]);
+		free(reader->csts[0]);
+		for(int i = 0; i < reader->num_ugs; i++) {
+			recorder_free_cfg(reader->cfgs[i]);
+			free(reader->cfgs[i]);
+		}
+	}
+
+	free(reader->csts);
+	free(reader->cfgs);
+
+    memset(reader, 0, sizeof(*reader));
 }
 
 const char* recorder_get_func_name(RecorderReader* reader, Record* record) {
@@ -70,9 +131,9 @@ const char* recorder_get_func_name(RecorderReader* reader, Record* record) {
 }
 
 int recorder_get_func_type(RecorderReader* reader, Record* record) {
-    if(record->func_id < mpi_start_idx)
+    if(record->func_id < reader->mpi_start_idx)
         return RECORDER_POSIX;
-    if(record->func_id < hdf5_start_idx) {
+    if(record->func_id < reader->hdf5_start_idx) {
         const char* func_name = recorder_get_func_name(reader, record);
         if(strncmp(func_name, "MPI_File", 8) == 0)
             return RECORDER_MPIIO;
@@ -83,7 +144,12 @@ int recorder_get_func_type(RecorderReader* reader, Record* record) {
     return RECORDER_HDF5;
 }
 
-void cs_to_record(CallSignature *cs, Record *record) {
+// Caller needs to free the record after use
+// with the recorder_free_record() call.
+Record* recorder_cs_to_record(CallSignature *cs) {
+
+    Record *record = malloc(sizeof(Record));
+
     char* key = cs->key;
 
     int pos = 0;
@@ -113,12 +179,13 @@ void cs_to_record(CallSignature *cs, Record *record) {
     }
 
     assert(ai == record->arg_count);
+    return record;
 }
 
 void recorder_free_cst(CST* cst) {
     for(int i = 0; i < cst->entries; i++)
-        free(cst->cst_list[i].key);
-    free(cst->cst_list);
+        free(cst->cs_list[i].key);
+    free(cst->cs_list);
 }
 
 void recorder_free_cfg(CFG* cfg) {
@@ -134,9 +201,13 @@ void recorder_free_record(Record* r) {
     for(int i = 0; i < r->arg_count; i++)
         free(r->args[i]);
     free(r->args);
+    free(r);
 }
 
-void recorder_read_cst(RecorderReader *reader, int rank, CST *cst) {
+void recorder_read_cst(RecorderReader *reader, int rank) {
+	reader->csts[rank] = malloc(sizeof(CST));
+	CST* cst = reader->csts[rank];
+
     cst->rank = rank;
     char cst_filename[1096] = {0};
     sprintf(cst_filename, "%s/%d.cst", reader->logs_dir, rank);
@@ -146,24 +217,32 @@ void recorder_read_cst(RecorderReader *reader, int rank, CST *cst) {
     int key_len;
     fread(&cst->entries, sizeof(int), 1, f);
 
-    cst->cst_list = malloc(cst->entries * sizeof(CallSignature));
+	// cst->cs_list will be stored in the terminal_id order.
+    cst->cs_list = malloc(cst->entries * sizeof(CallSignature));
 
     for(int i = 0; i < cst->entries; i++) {
-        fread(&cst->cst_list[i].terminal, sizeof(int), 1, f);
-        fread(&cst->cst_list[i].key_len, sizeof(int), 1, f);
+		int terminal_id;
+        fread(&terminal_id, sizeof(int), 1, f);
+		assert(terminal_id < cst->entries);
 
-        cst->cst_list[i].key = malloc(cst->cst_list[i].key_len);
-        fread(cst->cst_list[i].key, 1, cst->cst_list[i].key_len, f);
+		CallSignature* cs = &(cst->cs_list[terminal_id]);
+		cs->terminal_id = terminal_id;
 
-        assert(cst->cst_list[i].terminal < cst->entries);
+        fread(&(cs->rank), sizeof(int), 1, f);
+        fread(&(cs->key_len), sizeof(int), 1, f);
+        fread(&(cs->count), sizeof(int), 1, f);
+
+        cs->key = malloc(cs->key_len);
+        fread(cs->key, 1, cs->key_len, f);
     }
     fclose(f);
-
-    //for(int i = 0; i < cst->entries; i++)
-    //    printf("%d, terminal %d, key len: %d\n", i, cst->cst_list[i].terminal, cst->cst_list[i].key_len);
 }
 
-void recorder_read_cfg(RecorderReader *reader, int rank, CFG* cfg) {
+
+void recorder_read_cfg(RecorderReader *reader, int rank) {
+	reader->cfgs[rank] = malloc(sizeof(CFG));
+	CFG* cfg = reader->cfgs[rank];
+
     cfg->rank = rank;
     char cfg_filename[1096] = {0};
     sprintf(cfg_filename, "%s/%d.cfg", reader->logs_dir, rank);
@@ -187,14 +266,29 @@ void recorder_read_cfg(RecorderReader *reader, int rank, CFG* cfg) {
     fclose(f);
 }
 
+void recorder_get_cst_cfg(RecorderReader* reader, int rank, CST** cst, CFG** cfg) {
+	if(reader->metadata.interprocess_compression) {
+		// Do nothing, as we have already read
+		// csts and cfgs during initialization
+	} else {
+		if(reader->csts[rank] == NULL)
+			recorder_read_cst(reader, rank);
+		if(reader->cfgs[rank] == NULL)
+			recorder_read_cfg(reader, rank);
+	}
+
+	*cst = reader->csts[rank];
+	*cfg = reader->cfgs[reader->ug_ids[rank]];
+}
+
 
 #define TERMINAL_START_ID 0
 
-void rule_application(RecorderReader* reader, RuleHash* rules, int rule_id, CallSignature *cst_list, FILE* ts_file,
+void rule_application(RecorderReader* reader, CFG* cfg, CST* cst, int rule_id, FILE* ts_file,
                       void (*user_op)(Record*, void*), void* user_arg, int free_record) {
 
     RuleHash *rule = NULL;
-    HASH_FIND_INT(rules, &rule_id, rule);
+    HASH_FIND_INT(cfg->cfg_head, &rule_id, rule);
     assert(rule != NULL);
 
     for(int i = 0; i < rule->symbols; i++) {
@@ -202,24 +296,23 @@ void rule_application(RecorderReader* reader, RuleHash* rules, int rule_id, Call
         int sym_exp = rule->rule_body[2*i+1];
         if (sym_val >= TERMINAL_START_ID) { // terminal
             for(int j = 0; j < sym_exp; j++) {
-                Record record;
-                cs_to_record(&cst_list[sym_val], &record);
+                Record* record = recorder_cs_to_record(&(cst->cs_list[sym_val]));
 
                 // Fill in timestamps
                 uint32_t ts[2];
                 fread(ts, sizeof(uint32_t), 2, ts_file);
-                record.tstart = ts[0] * reader->metadata.time_resolution + prev_tstart;
-                record.tend   = ts[1] * reader->metadata.time_resolution + prev_tstart;
-                prev_tstart = record.tstart;
+                record->tstart = ts[0] * reader->metadata.time_resolution + reader->prev_tstart;
+                record->tend   = ts[1] * reader->metadata.time_resolution + reader->prev_tstart;
+                reader->prev_tstart = record->tstart;
 
-                user_op(&record, user_arg);
+                user_op(record, user_arg);
 
                 if(free_record)
-                    recorder_free_record(&record);
+                    recorder_free_record(record);
             }
         } else {                            // non-terminal (i.e., rule)
             for(int j = 0; j < sym_exp; j++)
-                rule_application(reader, rules, sym_val, cst_list, ts_file, user_op, user_arg, free_record);
+                rule_application(reader, cfg, cst, sym_val, ts_file, user_op, user_arg, free_record);
         }
     }
 }
@@ -227,29 +320,38 @@ void rule_application(RecorderReader* reader, RuleHash* rules, int rule_id, Call
 
 // Decode all records for one rank
 // one record at a time
-void recorder_decode_records(RecorderReader *reader, CST *cst, CFG *cfg,
-                             void (*user_op)(Record*, void*), void* user_arg) {
-    assert(cst->rank == cfg->rank);
+void recorder_decode_records_core(RecorderReader *reader, CST *cst, CFG *cfg,
+                             void (*user_op)(Record*, void*), void* user_arg, bool free_record) {
 
-    prev_tstart = 0;
+    reader->prev_tstart = 0.0;
 
     char ts_filename[1096] = {0};
     sprintf(ts_filename, "%s/%d.ts", reader->logs_dir, cst->rank);
     FILE* ts_file = fopen(ts_filename, "rb");
 
-    rule_application(reader, cfg->cfg_head, -1, cst->cst_list, ts_file, user_op, user_arg, true);
+    rule_application(reader, cfg, cst, -1, ts_file, user_op, user_arg, free_record);
 
     fclose(ts_file);
 }
 
+void recorder_decode_records(RecorderReader *reader, int rank,
+                             void (*user_op)(Record*, void*), void* user_arg) {
+
+	CST* cst;
+	CFG* cfg;
+	recorder_get_cst_cfg(reader, rank, &cst, &cfg);
+    recorder_decode_records_core(reader, cst, cfg, user_op, user_arg, true);
+}
+
+
 
 /**
  * Similar to rule application, but only calcuate
  * the total number of calls if uncompressed.
  */
-size_t get_uncompressed_count(RecorderReader* reader, RuleHash* rules, int rule_id) {
+size_t get_uncompressed_count(RecorderReader* reader, CFG* cfg, int rule_id) {
     RuleHash *rule = NULL;
-    HASH_FIND_INT(rules, &rule_id, rule);
+    HASH_FIND_INT(cfg->cfg_head, &rule_id, rule);
     assert(rule != NULL);
 
     size_t count = 0;
@@ -260,7 +362,7 @@ size_t get_uncompressed_count(RecorderReader* reader, RuleHash* rules, int rule_
         if (sym_val >= TERMINAL_START_ID) { // terminal
             count += sym_exp;
         } else {                            // non-terminal (i.e., rule)
-            count += sym_exp * get_uncompressed_count(reader, rules, sym_val);
+            count += sym_exp * get_uncompressed_count(reader, cfg, sym_val);
         }
     }
 
@@ -269,6 +371,9 @@ size_t get_uncompressed_count(RecorderReader* reader, RuleHash* rules, int rule_
 
 
 
+/**
+ * Code below is used for recorder-viz
+ */
 typedef struct records_with_idx {
     PyRecord* records;
     int idx;
@@ -287,6 +392,9 @@ void insert_one_record(Record *record, void* arg) {
     r->args = record->args;
 
     ri->idx++;
+
+    // free record but not record->args
+    free(record);
 }
 
 PyRecord** read_all_records(char* traces_dir, size_t* counts) {
@@ -297,30 +405,19 @@ PyRecord** read_all_records(char* traces_dir, size_t* counts) {
     PyRecord** records = malloc(sizeof(PyRecord*) * reader.metadata.total_ranks);
 
     for(int rank = 0; rank < reader.metadata.total_ranks; rank++) {
-        CST cst;
-        CFG cfg;
-        recorder_read_cst(&reader, rank, &cst);
-        recorder_read_cfg(&reader, rank, &cfg);
 
-        counts[rank] = get_uncompressed_count(&reader, cfg.cfg_head, -1);
+        CST* cst;
+        CFG* cfg;
+		recorder_get_cst_cfg(&reader, rank, &cst, &cfg);
+
+        counts[rank] = get_uncompressed_count(&reader, cfg, -1);
         records[rank] = malloc(sizeof(PyRecord)* counts[rank]);
 
         records_with_idx_t ri;
         ri.records = records[rank];
         ri.idx = 0;
 
-        // From recorder_decode_records() but does not free the record
-        // ------------
-        prev_tstart = 0;
-        char ts_filename[1096] = {0};
-        sprintf(ts_filename, "%s/%d.ts", reader.logs_dir, cst.rank);
-        FILE* ts_file = fopen(ts_filename, "rb");
-        rule_application(&reader, cfg.cfg_head, -1, cst.cst_list, ts_file, insert_one_record, &ri, false);
-        fclose(ts_file);
-        // ------------
-
-        recorder_free_cst(&cst);
-        recorder_free_cfg(&cfg);
+        recorder_decode_records_core(&reader, cst, cfg, insert_one_record, &ri, false);
     }
 
     recorder_free_reader(&reader);
